module dmd.OnScopeStatement;

import dmd.Statement;
import dmd.OutBuffer;
import dmd.HdrGenState;
import dmd.Scope;
import dmd.IRState;
import dmd.TOK;
import dmd.Loc;
import dmd.BE;
import dmd.Identifier;
import dmd.ExpInitializer;
import dmd.IntegerExp;
import dmd.VarDeclaration;
import dmd.Type;
import dmd.AssignExp;
import dmd.VarExp;
import dmd.NotExp;
import dmd.IfStatement;
import dmd.DeclarationStatement;
import dmd.ExpStatement;
import dmd.Expression;
import dmd.Lexer;

class OnScopeStatement : Statement
{
    TOK tok;
    Statement statement;

    this(Loc loc, TOK tok, Statement statement)
	{
		super(loc);
		
		this.tok = tok;
		this.statement = statement;
	}

    Statement syntaxCopy()
	{
		assert(false);
	}

    BE blockExit()
	{
		// At this point, this statement is just an empty placeholder
		return BE.BEfallthru;
	}

    void toCBuffer(OutBuffer buf, HdrGenState* hgs)
	{
		assert(false);
	}

    Statement semantic(Scope sc)
	{
		/* semantic is called on results of scopeCode() */
		return this;
	}

    bool usesEH()
	{
		assert(false);
	}

    void scopeCode(Scope sc, Statement* sentry, Statement* sexception, Statement* sfinally)
	{
		//printf("OnScopeStatement::scopeCode()\n");
		//print();
		*sentry = null;
		*sexception = null;
		*sfinally = null;
		switch (tok)
		{
			case TOKon_scope_exit:
				*sfinally = statement;
				break;

			case TOKon_scope_failure:
				*sexception = statement;
				break;

			case TOKon_scope_success:
			{
				/* Create:
				 *	sentry:   int x = 0;
				 *	sexception:    x = 1;
				 *	sfinally: if (!x) statement;
				 */
				Identifier id = Lexer.uniqueId("__os");

				ExpInitializer ie = new ExpInitializer(loc, new IntegerExp(0));
				VarDeclaration v = new VarDeclaration(loc, Type.tint32, id, ie);
				*sentry = new DeclarationStatement(loc, v);

				Expression e = new IntegerExp(1);
				e = new AssignExp(Loc(0), new VarExp(Loc(0), v), e);
				*sexception = new ExpStatement(Loc(0), e);

				e = new VarExp(Loc(0), v);
				e = new NotExp(Loc(0), e);
				*sfinally = new IfStatement(Loc(0), null, e, statement, null);

				break;
			}

			default:
				assert(0);
		}
	}

    void toIR(IRState* irs)
	{
	}
}